# Amethyst - An experimental Minecraft server written in Elixir.
# Copyright (C) 2024 KodiCraft
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

defmodule Amethyst.ConnectionHandler do
  @moduledoc """
  This module is responsible for handling incoming packets and sending outgoing packets. It keeps track of what state the connection is in and which game should
  receive the packets.
  """
alias ElixirSense.Log
  require Logger

  @spec child_spec(:gen_tcp.socket()) :: Supervisor.child_spec()
  def child_spec(socket) do
    %{
      id: __MODULE__,
      start: {__MODULE__, :start, [socket, Amethyst.ConnectionState.Handshake, 0]}
    }
  end

  @spec start(:gen_tcp.socket(), atom(), pos_integer()) :: no_return()
  def start(socket, connstate, version) do
    {:ok, spawn(fn ->
      Process.set_label("ConnectionHandler for #{inspect(socket)}")
      loop(socket, connstate, version, %{})
    end)}
  end

  @spec start_link(:gen_tcp.socket(), atom(), pos_integer()) :: no_return()
  def start_link(socket, connstate, version) do
    {:ok, spawn_link(fn ->
      Process.set_label("ConnectionHandler for #{inspect(socket)}")
      loop(socket, connstate, version, %{})
    end)}
  end

  @spec loop(:gen_tcp.socket(), atom(), pos_integer(), map()) :: no_return()
  defp loop(socket, connstate, version, state) do
    receive do
      :closed ->
        Logger.info("Connection #{inspect(socket)} closed.")
        Process.exit(self(), :normal)
      {:disconnect, reason} ->
        disconnect(socket, reason, connstate, version)
        Process.exit(self(), :normal)
      {:set_state, newstate} ->
        Logger.debug("Switching to state #{newstate} from #{connstate}")
        loop(socket, newstate, version, state)
      {:set_version, newversion} ->
        Logger.debug("Switching to version #{newversion} from #{version}")
        loop(socket, connstate, newversion, state)
      {:send_packet, packet} ->
        Logger.debug("Sending packet #{inspect(packet)}")
        send_packet(socket, connstate, packet, version)
        loop(socket, connstate, version, state)
      after 0 ->
        receive do
          {:packet, id, data} ->
            state = handle_packet(id, data, connstate, version, state)
            loop(socket, connstate, version, state)
        end
    end
  end

  defp handle_packet(id, data, connstate, version, state) do
    try do
      packet = connstate.deserialize(id, version, data)
      case connstate.handle(packet, version, state) do
        :ok -> state
        {:error, reason} ->
          Logger.error("Error handling packet with ID #{id} in state #{connstate}: #{reason}")
          send(self(), {:disconnect, "Error handling packet #{id}:\n#{reason}"})
          state
        newstate ->
          if is_map(newstate) do
            newstate
          else
            Logger.warning("State change to #{newstate} is not a map! Did you forget to return :ok?")
            state
          end
      end
    rescue
      e ->
        Logger.error("Error handling packet with ID #{id} in state #{connstate}: #{Exception.format(:error, e, __STACKTRACE__)}")
        send(self(), {:disconnect, "Error handling packet #{id}:\n#{e.message}"})
        state
    end
  end

  defp send_packet(socket, connstate, packet, version) do
    data = connstate.serialize(packet, version)
    length = byte_size(data) |> Amethyst.Minecraft.Write.varint()
    :gen_tcp.send(socket, length <> data)
  end

  defp disconnect(socket, reason, connstate, version) do
    Logger.info("Disconnecting connection #{inspect(socket)}")
    Logger.debug("Disconnecting connection #{inspect(socket)}: #{reason}")
    case connstate.disconnect(reason) do
      nil -> nil
      packet -> send_packet(socket, connstate, packet, version)
    end
    :gen_tcp.close(socket)
  end
end
