# Amethyst - An experimental Minecraft server written in Elixir.
# Copyright (C) 2024 KodiCraft
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

defmodule Amethyst.ConnectionReceiver do
  @moduledoc """
  This module waits for data incoming through a TCP connection, reads entire packets at a time and sends them to the handler.
  """

  require Logger
  alias Amethyst.Minecraft.Read

  @spec child_spec(:gen_tcp.socket()) :: Supervisor.child_spec()
  def child_spec(socket) do
    %{
      id: __MODULE__,
      start: {__MODULE__, :start, [socket]}
    }
  end

  @spec start(:gen_tcp.socket()) :: no_return()
  def start(socket) do
    {:ok, spawn(fn ->
      Process.set_label("ConnectionReceiver for #{inspect(socket)}")
      {:ok, pid} = Amethyst.ConnectionHandler.start_link(socket, Amethyst.ConnectionState.Handshake, 0)
      receive(socket, pid)
    end)}
  end

  @spec start_link(:gen_tcp.socket()) :: no_return()
  def start_link(socket) do
    {:ok, spawn_link(fn ->
      Process.set_label("ConnectionReceiver for #{inspect(socket)}")
      {:ok, pid} = Amethyst.ConnectionHandler.start_link(socket, Amethyst.ConnectionState.Handshake, 0)
      receive(socket, pid)
    end)}
  end

  @spec receive(:gen_tcp.socket(), pid()) :: no_return()
  def receive(socket, sender) do
    case get_packet(socket) do
      :closed -> send(sender, :closed)
        Process.exit(self(), :normal)
      {:error, error} -> Logger.error("Error reading packet: #{error}")
      {id, data} -> send(sender, {:packet, id, data})
    end
    receive(socket, sender)
  end

  def get_packet(client) do
    case get_varint(client, "") do
      :closed -> :closed
      {:error, error} -> {:error, error}
      {[length], ""} ->
        recv = :gen_tcp.recv(client, length)
        case recv do
          {:ok, full_packet} -> ({[id], data} = Read.start(full_packet) |> Read.varint() |> Read.stop()
                                {id, data})
          {:error, :closed} -> :closed
          {:error, error} -> {:error, error}
        end
    end
  end

  defp get_varint(client, acc) do
    case :gen_tcp.recv(client, 1) do
      {:ok, byte} -> case byte do
        <<0::1, _::7>> -> Read.start(acc <> byte) |> Read.varint() |> Read.stop()
        <<1::1, _::7>> -> get_varint(client, acc <> byte)
      end
      {:error, :closed} -> :closed
      {:error, error} -> {:error, error}
    end
  end
end
